/* owwl.c */

/* oww_trx client library */

/* Dr. Simon J. Melhuish */


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#ifndef OWWL_DEBUG
#define NDEBUG 1
#endif /* ifdef OWWL_DEBUG */


#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include <math.h>
#include <time.h>
#include <assert.h>

#include "sockdef.h"

#ifdef VXW
#include <ioLib.h> /* Needed for FIONBIO */
#include <selectLib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* ifdef HAVE_SYS_TYPES_H */

#ifdef HAVE_INTTYPES_H
# include <inttypes.h>
#endif

#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif /* HAVE_SYS_TYPES_H */

#include "owwl.h"

static void owwl_tx_server_kill_clients(owwl_server *server) ;

long owwl_remote_version ;

// static int Owwl_MaxArg[] = {
//   0, 0, 0, 2, 3, 4
// } ;

/* Unit strings */

char *OwwlUnitName_Temperature[] = {
  "\xc2\xb0""C", "\xc2\xb0""F", "K", "R"
} ;

char *OwwlUnitName_Length[] = {
  "mm", "\"", "mm", "\""
} ;

char *OwwlUnitName_LengthRate[] = {
  "mm/h", "\"/h", "mm/h", "\"/h"
} ;

char *OwwlUnitName_Count[] = {
  "counts", "counts", "counts", "counts"
} ;

char *OwwlUnitName_CountRate[] = {
  "counts/s", "counts/h", "counts/s", "counts/h"
} ;

char *OwwlUnitName_Speed[] = {
  "kph", "mph", "m/s", "Knots"
} ;

char *OwwlUnitName_Pressure[] = {
  "mBar", "\"Hg", "kPa", "atm"
} ;

char *OwwlUnitName_Direction[] = {
  "\xc2\xb0", "", "", "rad"
} ;

char *OwwlUnitDesc_Direction[] = {
  "\xc2\xb0", "Point No.", "Compass Point", "Radians"
} ;

char *OwwlUnitName_Humidity[] = {
  "%", "%", "%", "%"
} ;

char *OwwlUnitName_Uvi[] = {
	/* The index is dimensionless, but we can convert to flux */
  "", "", "mW/m\xc2\xb2", "mW/m\xc2\xb2"
} ;

char *OwwlUnitName_Time[] = {
  "", "", "", ""
} ;

char *OwwlUnitName_Volts[] = {
  "V", "mV", "V", "mV"
} ;

char *OwwlUnitName_mVolts[] = {
  "mV", "µV", "mV", "µV"
} ;

char *OwwlUnitName_mAmps[] = {
	"mA", "mA", "mA", "mA"
} ;

char *OwwlUnitName_MicroVolts[] = {
  "µV", "µV", "µV", "µV"
} ;

char *OwwlUnitName_NanoAmps[] = {
  "nA", "nA", "nA", "nA"
} ;

char *OwwlUnitName_Coulombs[] = {
  "C", "C", "C", "C"
} ;

char *OwwlUnitName_mVs[] = {
  "mV.s", "mV.hr", "V.s", "V.hr"
} ;

char *OwwlUnitName_Ohms[] = {
  "\xe2\x84\xa6", "k\xe2\x84\xa6", "\xe2\x84\xa6", "k\xe2\x84\xa6"
} ;

char *OwwlUnitName_Cryo[] = {
  "K", "R", "\xc2\xb0""C", "\xc2\xb0""F"
} ;

char *OwwlUnitName_Dimensionless[] = {
  "", "", "", ""
} ;

char *OwwlUnitDesc_Time[] = {
  "Time and local date",
  "Time only",
  "ISO 8601 long form",
  "ISO 8601 short form"
} ;

char *OwwlUnitFormat_Time[] = {
  /* Time and local date */ "%H:%M %x",
  /* Time only */           "%H:%M",
  /* ISO 8601 long form */  "%Y-%m-%d %H:%M",
  /* ISO 8601 short form */ "%y%m%d %H%M"
} ;

char **OwwlUnitNames[] = {
  OwwlUnitName_Dimensionless,
  OwwlUnitName_Time,
  OwwlUnitName_Count,
  OwwlUnitName_CountRate,
  OwwlUnitName_Length,
  OwwlUnitName_LengthRate,
  OwwlUnitName_Speed,
  OwwlUnitName_Temperature,
  OwwlUnitName_Pressure,
  OwwlUnitName_Direction,
  OwwlUnitName_Humidity,
  OwwlUnitName_Uvi,
  OwwlUnitName_Volts,
  OwwlUnitName_mVolts,
  OwwlUnitName_MicroVolts,
  OwwlUnitName_NanoAmps,
  OwwlUnitName_Cryo,
  OwwlUnitName_Ohms,
  OwwlUnitName_mAmps,
  OwwlUnitName_Coulombs,
  OwwlUnitName_mVs
} ;

char **OwwlUnitStrings[] = {
  OwwlUnitName_Dimensionless,
  OwwlUnitDesc_Time,
  OwwlUnitName_Count,
  OwwlUnitName_CountRate,
  OwwlUnitName_Length,
  OwwlUnitName_LengthRate,
  OwwlUnitName_Speed,
  OwwlUnitName_Temperature,
  OwwlUnitName_Pressure,
  OwwlUnitDesc_Direction,
  OwwlUnitName_Humidity,
  OwwlUnitName_Uvi,
  OwwlUnitName_Volts,
  OwwlUnitName_mVolts,
  OwwlUnitName_MicroVolts,
  OwwlUnitName_NanoAmps,
  OwwlUnitName_Cryo,
  OwwlUnitName_Ohms,
  OwwlUnitName_mAmps,
  OwwlUnitName_Coulombs,
  OwwlUnitName_mVs
} ;

/* Get the unit class for an arg */

enum owwl_unit_class_enum owwl_unit_class(owwl_data *data, int arg)
{
  /* Switch by device type */
  switch (data->device_type)
  {
    case OwwlDev_Temperature: /* Normal thermometer */
    {
      return owwl_unit_class_temperature ;
    }

    case OwwlDev_Barometer:
    {
      return owwl_unit_class_pressure ;
    }

    case OwwlDev_Humidity:
    {
      return owwl_unit_class_humidity ;
    }

    case OwwlDev_Wind:
    {
      switch(arg)
      {
        case 0: /* Speed */
        case 1: /* Gust */
          return  owwl_unit_class_speed;

        case 2: /* Bearing */
          return owwl_unit_class_direction ;

        default:
          return owwl_unit_class_unknown ;
      }
    }

    case OwwlDev_Rain:
    {
      switch(arg)
      {
        case 0: /* Rain since reset */
          return owwl_unit_class_length ;

        case 1: /* Reset time */
          return owwl_unit_class_time ;

        case 2: /* Rain rate */
          return owwl_unit_class_length_rate ;

        case 3: /* Counts */
          return  owwl_unit_class_count ;

        default:
          return owwl_unit_class_unknown ;
      }
    }

    case OwwlDev_Gpc:
    {
      switch(arg)
      {
        case 0: /* Counts since reset */
        case 3: /* Counts */
        case 4: /* Delta */
          return owwl_unit_class_count ;

        case 1: /* Reset time */
          return owwl_unit_class_time ;

        case 2: /* Count rate */
          return owwl_unit_class_count_rate ;

        default:
          return owwl_unit_class_unknown ;
      }
    }

    case OwwlDev_Uvi:
      return owwl_unit_class_uvi ;

    case OwwlDev_Adc:
      switch (arg)
      {
      case 0:
    	return owwl_unit_class_voltage;

      case 1:
    	return owwl_unit_class_mvoltage;

      case 2:
    	return owwl_unit_class_mvs;

      default:
    	return owwl_unit_class_unknown ;
      }
      break;

    case OwwlDev_SoilMoisture:
    	return owwl_unit_class_pressure;

    case OwwlDev_Solar:
    case OwwlDev_LeafWetness:
    case OwwlDev_Psb:
      return owwl_unit_class_dimensionless ;

    case OwwlDev_Diode:
      return owwl_unit_class_voltage ;

    case OwwlDev_Heater:
    {
      switch (arg)
      {
      case 0: /* Voltage */
        return owwl_unit_class_voltage ;
    	
      case 1: /* Isense */
        return owwl_unit_class_mamps ;
      }
    }

    case OwwlDev_Resistance:
      return owwl_unit_class_resistance ;

    case OwwlDev_Grt:
    {
      switch(arg)
      {
        case 0: /* V */
        case 1: /* Vq */
          return owwl_unit_class_muvoltage ;

        case 2: /* I */
          return owwl_unit_class_nanocurrent ;

        default:
          return owwl_unit_class_unknown ;
      }
    }

    case OwwlDev_Cryo: /* Cryogenic calibrated thermometer - internal representation is K */
      return owwl_unit_class_cryo ;

    default:
      return owwl_unit_class_unknown ;
  }

  return owwl_unit_class_unknown ;
}

/* Get the unit name string for a data value */
char *owwl_unit_name(owwl_data *data, int unit_type, int arg)
{
  enum owwl_unit_class_enum unit_class ;
  unit_class = owwl_unit_class(data, arg) ;
  if (unit_class < 0) unit_class = 0 ;
  return OwwlUnitNames[unit_class][unit_type] ;
}

/* Get the unit description string for a data value */
char *owwl_unit_desc(owwl_data *data, int unit_type, int arg)
{
  enum owwl_unit_class_enum unit_class ;
  unit_class = owwl_unit_class(data, arg) ;
  if (unit_class < 0) unit_class = 0 ;
  return OwwlUnitStrings[unit_class][unit_type] ;
}

static double
convert_temperature(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_Celcius:
      return val ;

    case OwwlUnit_Fahrenheit:
      return 32.0 + 1.8 * val ;

    case OwwlUnit_Kelvin:
      return val + 273.15 ;

    case OwwlUnit_Rankine:
      return 1.8 * (val + 273.15) ;
  }

  return 0.0 ;
}

static double
readout_temperature(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Temperature)) ;

  if (arg != 0) return 0.0 ;

  return (convert_temperature(unit_type, self->device_data.temperature.T)) ;
}

static char stringbuff[256] ;

static char *
string_from_temperature(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1) prec = 1 ;

  sprintf(stringbuff, "%.*f", prec, readout_temperature(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_pressure(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_Mbar:
      return val ;

    case OwwlUnit_InchHg:
      return 0.0295299833 * val ;

    case OwwlUnit_Kpa:
      return 0.1 * val ;

    case OwwlUnit_Atm:
      return 0.000986923267 * val ;
  }

  return 0.0 ;
}

static double
readout_barometer(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Barometer)) ;

  if (arg != 0) return 0.0 ;
  return convert_pressure(unit_type, self->device_data.barom.bp) ;
}

static char *
string_from_barometer(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  int precs[] = {1, 3, 2, 4} ;
  double bp;

  assert((unit_type >= 0)) ;
  assert((unit_type < 4)) ;
  
  bp = readout_barometer(self, unit_type, arg);


  if (bp < 100.0)
  {
    if (prec == -1) prec = 5 ;
    sprintf(stringbuff, "%.*g", prec, bp) ;
  }
  else
  {
    if (prec == -1) prec = precs[unit_type] ;
    sprintf(stringbuff, "%.*f", prec, bp) ;
  }

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_humidity(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Humidity)) ;

  if (arg != 0) return 0.0 ;
  return (double) self->device_data.humidity.RH ;
}

static char *
string_from_humidity(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 1 ;

  sprintf(stringbuff, "%.*f", prec, readout_humidity(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_length(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_Mm:
    case OwwlUnit_Alt1:
      return val ;

    case OwwlUnit_Inch:
    case OwwlUnit_Alt2:
      return val / 25.4 ;
  }

  return 0.0 ;
}

static double
readout_rain(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Rain)) ;

  switch (arg)
  {
    case 0: /* Rain since reset */
      return convert_length(unit_type, self->device_data.rain.rain_since_reset) ;
    break ;

    case 1: /* Reset date / time */
      return (double) self->device_data.rain.rain_reset_time ;
    break ;

    case 2: /* Rain rate */
      return convert_length(unit_type, self->device_data.rain.rain_rate) ;
    break ;

    case 3: /* Monotonic count */
      return (double) self->device_data.rain.rain_count ;
    break ;
  }

  return 0.0 ;
}

static char *
string_from_rain(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  assert((unit_type >= 0)) ;
  assert((unit_type < 4)) ;

  switch (arg)
  {
    case 0: /* Rain since reset */
    case 2: /* Rain rate */
    {
      int precs[] = {1, 2, 1, 2} ;
      if (prec == -1)
        prec = precs[unit_type] ;

      sprintf(stringbuff, "%.*f", prec, readout_rain(self, unit_type, arg)) ;
    }
    break ;

    case 1: /* Reset date / time */
      /*strncpy(buffer, ctime(&self->device_data.rain.rain_reset_time), 63) ; */
	    strftime(stringbuff, sizeof(stringbuff)-1,
	      OwwlUnitFormat_Time[unit_type], localtime(&self->device_data.rain.rain_reset_time)) ;
    break ;

    case 3: /* Monotonic count */
      sprintf(stringbuff, "%u", self->device_data.rain.rain_count) ;
    break ;
  }

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_rate(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_CountsPerSec:
    case OwwlUnit_Alt1:
      return val ;

    case OwwlUnit_CountsPerHour:
    case OwwlUnit_Alt2:
      return val * 3600.0 ;
  }

  return 0.0 ;
}

static double
readout_gpc(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Gpc)) ;

  switch (arg)
  {
    case 0: /* GPC since reset */
      return self->device_data.gpc.gpc_calibrated ;
    break ;

    case 1: /* Reset date / time */
      return (double) self->device_data.gpc.gpc_reset_time ;
    break ;

    case 2: /* GPC rate */
      return convert_rate(unit_type, self->device_data.gpc.gpc_rate);
    break ;

    case 3: /* Monotonic count */
      return (double) self->device_data.gpc.gpc_count ;
    break ;

    case 4: /* GPC delta */
      return (double) self->device_data.gpc.gpc_delta ;
    break ;
  }

  return 0.0 ;
}

static char *
string_from_gpc(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  assert((unit_type >= 0)) ;
  assert((unit_type < 4)) ;

  switch (arg)
  {
    case 0: /* GPC since reset */
    case 2: /* GPC rate */
    case 4: /* GPC delta */
    {
      if (-1 == prec)
        sprintf(stringbuff, "%f", readout_gpc(self, unit_type, arg)) ;
      else
        sprintf(stringbuff, "%.*f", prec, readout_gpc(self, unit_type, arg)) ;
    }
    break ;

    case 1: /* Reset date / time */
      /*strncpy(stringbuff, ctime(&self->device_data.gpc.gpc_reset_time), 63) ; */
	    strftime(stringbuff, sizeof(stringbuff)-1,
	      "%H:%M %x", localtime(&self->device_data.gpc.gpc_reset_time)) ;
    break ;

    case 3: /* Monotonic count */
      sprintf(stringbuff, "%u", self->device_data.gpc.gpc_count) ;
    break ;
  }

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_solar(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Solar)) ;

  if (arg != 0) return 0.0 ;
  return (double) self->device_data.solar.solar ;
}

static char *
string_from_solar(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 1 ;

  sprintf(stringbuff, "%.*f", prec, readout_solar(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_uvi(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_UVI:
    case OwwlUnit_Imperial:
      return val ;

    case OwwlUnit_mWpersqm:
    case OwwlUnit_Alt2:
      return val * 10.0 ;
  }

  return 0.0 ;
}

static double
readout_uvi(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Uvi)) ;

  if (arg != 0) return 0.0 ;

  return (convert_uvi(unit_type, self->device_data.uvi.uvi)) ;
}

static char *
string_from_uvi(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = (unit_type >= OwwlUnit_mWpersqm) ? 0 : 1 ;

  sprintf(stringbuff, "%.*f", prec, readout_uvi(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_volts(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_V:
    case OwwlUnit_Alt1:
      return val ;

    case OwwlUnit_mV:
    case OwwlUnit_Alt2:
      return val * 1000.0F ;
  }

  return 0.0 ;
}

static double
convert_mvolts(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_mV_mV:
    case OwwlUnit_Alt1:
      return val ; // Unit is mV

    case OwwlUnit_mV_uV:
    case OwwlUnit_Alt2:
      return val * 1000.0F ; // Convert to uV
  }

  return 0.0 ;
}

static double
convert_mvoltsec(int unit_type, double val)
{
  /* Value is in mV.s */
  switch (unit_type)
  {
    case OwwlUnit_mVs_mVs:
      return val ; // Unit is mV.s

    case OwwlUnit_mVs_mVhr:
      return val / 3600.0 ; // Convert to mV.hr

    case OwwlUnit_mVs_Vs:
      return val * 0.001F ; // Convert to V.s

    case OwwlUnit_mVs_Vhr:
      return val / 3.6e6 ;  // Convert to V.hr
  }

  return 0.0 ;
}

static double
readout_adc(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Adc)) ;

  switch (arg)
  {
    case 0: /* Vdd */
      return convert_volts(unit_type, self->device_data.adc.V) ;

    case 1: /* Vss */
      return convert_mvolts(unit_type, self->device_data.adc.I) ;

    case 2: /* Vacc */
      return convert_mvoltsec(unit_type, self->device_data.adc.Q) ;
  }

  return 0.0 ;
}

static char *
string_from_adc(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
  {
	switch (arg)
	{
	case 0:
	  prec = 3 ;
	  break;

	case 1:
	  prec = 3 ;
	  break;

	case 2:
	{
	  switch (unit_type)
	  {
	  case OwwlUnit_mVs_mVs:
		prec =  0 ; // Unit is mV.s
		break;

	  case OwwlUnit_mVs_mVhr:
		prec =  3 ; // Unit is V.s
		break;

	  case OwwlUnit_mVs_Vs:
		prec =  3 ; // Unit is V.s
		break;

	  case OwwlUnit_mVs_Vhr:
		prec =  6 ; // Unit is V.s
		break;
	  }
	  break;
	}
	}
  }

//  printf("Arg %d prec %d unit %d\n", arg, prec, unit_type);

  sprintf(stringbuff, "%.*f", prec, readout_adc(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_soilmoisture(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_SoilMoisture)) ;

  if (arg != 0) return 0.0 ;
  return convert_pressure(unit_type, self->device_data.soilmoisture.soilmoisture) ;
}

static char *
string_from_soilmoisture(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 0 ;

  sprintf(stringbuff, "%.*f", prec, readout_soilmoisture(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_leafwetness(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_LeafWetness)) ;

  if (arg != 0) return 0.0 ;
  return (double) self->device_data.leafwetness.leafwetness ;
}

static char *
string_from_leafwetness(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 0 ;

  sprintf(stringbuff, "%.*f", prec, readout_leafwetness(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static int
readout_psb_i(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Psb)) ;

  switch (arg)
  {
    case 0: /* PSB channel A */
      return self->device_data.psb.psb_a ;

    case 1: /* PSB channel B */
      return self->device_data.psb.psb_b ;
  }

  return 0.0 ;
}

static double
readout_psb(owwl_data *self, int unit_type, int arg)
{
  return (double) readout_psb_i(self, unit_type, arg);
}

static char *
string_from_psb(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 1 ;

  sprintf(stringbuff, "%.*d", prec, readout_psb_i(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_diode(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Diode)) ;

  if (arg != 0) return 0.0 ;
  return (double) self->device_data.diode.V ;
}

static char *
string_from_diode(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 4 ;

  sprintf(stringbuff, "%.*f", prec, readout_diode(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_grt(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Grt)) ;

  switch (arg)
  {
    case 0: /* GRT V */
      return (double) self->device_data.grt.V ;

    case 1: /* GRT Vq */
      return (double) self->device_data.grt.Vq ;

    case 2: /* GRT I */
      return (double) self->device_data.grt.I ;
  }

  return 0.0 ;
}

static char *
string_from_grt(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 5 ;

  sprintf(stringbuff, "%#.*g", prec, readout_grt(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_cryo(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_CryoCelcius:
      return val - 273.15 ;

    case OwwlUnit_CryoFahrenheit:
      return 32.0 + 1.8 * (val - 273.15) ;

    case OwwlUnit_CryoKelvin:
      return val ;

    case OwwlUnit_CryoRankine:
      return 1.8 * val ;
  }

  return 0.0 ;
}

static double
readout_cryo(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Cryo)) ;

  if (arg != 0) return 0.0 ;

  return (convert_cryo(unit_type, self->device_data.temperature.T)) ;
}

static char *
string_from_cryo(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  double val ;

  val = readout_cryo(self, unit_type, arg) ;

  if (prec == -1)
  {
    switch (self->device_subtype)
    {
      // Give GRTs extra precision
      case OwwlCryo_Grt:
        if (val < 1.0)
          prec = 5 ;
        else if (val < 10.0)
          prec = 4 ;
        else if (val < 100.0)
          prec = 3 ;
        else
          prec = 2 ;
        break;
        
      default:
        if (val < 1.0)
          prec = 4 ;
        else if (val < 10.0)
          prec = 3 ;
        else
          prec = 2 ;
        break;
    }
  }

  sprintf(stringbuff, "%.*f", prec, val) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_heater(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Heater)) ;

  switch (arg) {
  case 0:
	return (double) self->device_data.heater.V ;

  case 1:
    return (double) self->device_data.heater.mA ;
  }
  
  return 0.0;
}

static char *
string_from_heater(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  if (prec == -1)
    prec = 4 ;

  sprintf(stringbuff, "%.*f", prec, readout_heater(self, unit_type, arg)) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
readout_resistance(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Resistance)) ;

  if (arg != 0) return 0.0 ;
  return (double) (unit_type%2==0) ? self->device_data.resistance.R : 0.001 * self->device_data.resistance.R ;
}

static char *
string_from_resistance(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  double R;
  R = readout_resistance(self, unit_type, arg);
  if (prec == -1)
  {
    if (unit_type%2==0) // Ohms
    {
      if (R < 102.4)  prec = 3 ;
      else if (R < 1024.0) prec = 2 ;
      else if (R < 10240.0) prec = 1 ;
      else prec = 0 ;
    }
    else // kOhms
   {
      if (R < 0.1024)  prec = 6 ;
      else if (R < 1.024) prec = 5 ;
      else if (R < 10.24) prec = 4 ;
      else if (R < 102.4) prec = 3 ;
      else if (R < 1024.0) prec = 2 ;
      else if (R < 10240.0) prec = 1 ;
      else prec = 0 ;
    }
  }

  sprintf(stringbuff, "%.*f", prec, R) ;

  return strncpy(buffer, stringbuff, bufflen) ;
}

static double
convert_speed(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_Kph:
      return 3.6 * val ;

    case OwwlUnit_Mph:
      return 2.23693629 * val ;

    case OwwlUnit_Mps:
      return val ;

    case OwwlUnit_Knot:
      return 1.94384449 * val ;
  }

  return 0.0 ;
}

static double
convert_bearing(int unit_type, double val)
{
  switch (unit_type)
  {
    case OwwlUnit_Bearing:
      return val ;

    case OwwlUnit_Point16:
    case OwwlUnit_Name:
      return 1.0 + val / 22.5 ;

    case OwwlUnit_Radians:
      return val * 0.017453292 ;
  }

  return 0.0 ;
}

static double
readout_wind(owwl_data *self, int unit_type, int arg)
{
  assert((self->device_type == OwwlDev_Wind)) ;

  switch (arg)
  {
    case 0: /* Speed */
      return convert_speed(unit_type, self->device_data.wind.speed) ;
    break ;

    case 1: /* Gust */
      return convert_speed(unit_type, self->device_data.wind.gust) ;
    break ;

    case 2: /* Bearing */
      return convert_bearing(unit_type, self->device_data.wind.bearing) ;
    break ;
  }

  return 0.0 ;
}

static char *
bearing_name(double val)
{
  static char *names[16] = {
    "N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE",
    "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW",
  } ;

  int bn = (int)(floor(0.5 + val/22.5)) ;

  if ((bn < 0) || (bn >= 16)) bn = 0 ;

  return names[bn] ;
}

static char *
string_from_wind(owwl_data *self, char *buffer, int bufflen, int unit_type, int prec, int arg)
{
  switch (arg)
  {
    case 0: /* Speed */
    case 1: /* Gust */
    {
      if (prec == -1) prec = 1 ; /* Default precision */
      sprintf(stringbuff, "%.*f", prec, readout_wind(self, unit_type, arg)) ;
    }
    break ;

    case 2: /* Bearing */
    {
      switch (unit_type)
      {
        case OwwlUnit_Bearing:
        case OwwlUnit_Radians:
        {
          if (-1 == prec) prec = (unit_type == 0) ? 1 : 2 ;
          sprintf(stringbuff, "%.*f", prec, readout_wind(self, unit_type, arg)) ;
        }
        break ;

        case OwwlUnit_Point16:
        {
          sprintf(stringbuff, "%d", (int) (floor(0.5 + self->device_data.wind.bearing/22.5))) ;
        }
        break ;

        case OwwlUnit_Name:
        {
          strcpy(stringbuff, bearing_name(self->device_data.wind.bearing)) ;
        }
        break ;
      }
    }
    break ;
  }

  return strncpy(buffer, stringbuff, bufflen) ;
}

/* Add a new device at the end of the data array */

static owwl_data*
owwl_append_new_device(owwl_conn *conn, int device_type, int device_subtype, int index)
{
  assert((conn != NULL)) ;

  conn->data = (owwl_data *) realloc(conn->data, (conn->data_count+1) * sizeof(owwl_data)) ;
  if (!conn->data)
  {
    conn->data_count = 0 ;
    return NULL ; /* Error */
  }

  conn->data[conn->data_count].device_type    = device_type ;
  conn->data[conn->data_count].index          = index ;
  conn->data[conn->data_count].in_use         = 0 ;
  conn->data[conn->data_count].device_subtype = device_subtype ;
  conn->data[conn->data_count].user_data      = NULL ;

  /* Device-specific initialization */
  switch (device_type)
  {
    case OwwlDev_Temperature:
      conn->data[conn->data_count].str = string_from_temperature ;
      conn->data[conn->data_count].val = readout_temperature ;
      break ;

    case OwwlDev_Barometer:
      conn->data[conn->data_count].str = string_from_barometer ;
      conn->data[conn->data_count].val = readout_barometer ;
      break ;

    case OwwlDev_Humidity:
      conn->data[conn->data_count].str = string_from_humidity ;
      conn->data[conn->data_count].val = readout_humidity ;
      break ;

    case OwwlDev_Wind:
      conn->data[conn->data_count].str = string_from_wind ;
      conn->data[conn->data_count].val = readout_wind ;
      break ;

    case OwwlDev_Rain:
      conn->data[conn->data_count].str = string_from_rain ;
      conn->data[conn->data_count].val = readout_rain ;
      break ;

    case OwwlDev_Gpc:
      conn->data[conn->data_count].str = string_from_gpc ;
      conn->data[conn->data_count].val = readout_gpc ;
      break ;

    case OwwlDev_Solar:
      conn->data[conn->data_count].str = string_from_solar ;
      conn->data[conn->data_count].val = readout_solar ;
      break ;

    case OwwlDev_Uvi:
      conn->data[conn->data_count].str = string_from_uvi ;
      conn->data[conn->data_count].val = readout_uvi ;
      break ;

    case OwwlDev_Adc:
      conn->data[conn->data_count].str = string_from_adc ;
      conn->data[conn->data_count].val = readout_adc ;
      break ;

    case OwwlDev_SoilMoisture:
      conn->data[conn->data_count].str = string_from_soilmoisture ;
      conn->data[conn->data_count].val = readout_soilmoisture ;
      break ;

    case OwwlDev_LeafWetness:
      conn->data[conn->data_count].str = string_from_leafwetness ;
      conn->data[conn->data_count].val = readout_leafwetness ;
      break ;

    case OwwlDev_Psb:
      conn->data[conn->data_count].str = string_from_psb ;
      conn->data[conn->data_count].val = readout_psb ;
      break ;

    case OwwlDev_Diode:
      conn->data[conn->data_count].str = string_from_diode ;
      conn->data[conn->data_count].val = readout_diode ;
      break ;

    case OwwlDev_Grt:
      conn->data[conn->data_count].str = string_from_grt ;
      conn->data[conn->data_count].val = readout_grt ;
      break ;

    case OwwlDev_Cryo:
      conn->data[conn->data_count].str = string_from_cryo ;
      conn->data[conn->data_count].val = readout_cryo ;
      break ;

    case OwwlDev_Heater:
      conn->data[conn->data_count].str = string_from_heater ;
      conn->data[conn->data_count].val = readout_heater ;
      break ;

    case OwwlDev_Resistance:
      conn->data[conn->data_count].str = string_from_resistance ;
      conn->data[conn->data_count].val = readout_resistance ;
      break ;

    default:
      conn->data[conn->data_count].str = NULL ;
      conn->data[conn->data_count].val = NULL ;
      break;
  }

  ++(conn->data_count) ;

  return (&(conn->data[conn->data_count-1])) ; /* Ok */
}

/* Try to find a device - create new device if not found */

static owwl_data*
owwl_ensure_device(owwl_conn *conn, int device_type, int device_subtype, int index)
{
  owwl_data *device ;

  device = owwl_find(conn, device_type, device_subtype, index) ;

  if (!device) device = owwl_append_new_device(conn, device_type, device_subtype, index) ;

  return device ;
}

/* Remove a device from the data array */

static int
owwl_remove_device(owwl_conn *conn, owwl_data *device)
{
  int i, skip = 0 ;

  assert((conn != NULL)) ;

  for (i=0; i < conn->data_count; ++i)
  {
    if ((!skip) && (&conn->data[i] == device))
    {
#     ifdef OWWL_DEBUG
        printf("Removing device %d\n", i) ;
#     endif /* ifdef OWWL_DEBUG */
      skip = 1 ;
    }

    if ((skip) && (i+1 < conn->data_count))
      conn->data[i] = conn->data[i+1] ;
  }

  --conn->data_count ;

  conn->data = (owwl_data *) realloc(conn->data, conn->data_count * sizeof(owwl_data)) ;

  return 0 ; /* Ok */
}

/* Remove all unused devices from data array */

static int
owwl_remove_unused_devices(owwl_conn *conn)
{
  int i ;

  assert((conn != NULL)) ;

  for (i=0; i < conn->data_count; ++i)
    if (!conn->data[i].in_use) owwl_remove_device(conn, &conn->data[i]) ;

  return 0 ; /* Ok */
}

static char *
owwl_name_stem_long(int devtype, int devsubtype)
{
  switch (devtype)
  {
    case OwwlDev_Temperature:
    {
      switch (devsubtype)
      {
        case OwwlTemp_Humidity:
          return "RH Thermometer" ;

        case OwwlTemp_Barometer:
          return "Bar. Thermometer" ;

        case OwwlTemp_UVI:
          return "UVI Thermometer" ;

        case OwwlTemp_Solar:
          return "Solar Thermometer" ;

        case OwwlTemp_Adc:
          return "ADC Thermometer" ;

        case OwwlTemp_Thrmcpl:
          return "Thermocouple" ;

        case OwwlTemp_ThrmcplCJ:
          return "Thermocouple cold jnctn" ;

        case OwwlTemp_Soil:
          return "Soil Thermometer";

        case OwwlTemp_Indoor:
          return "Indoor Thermometer";

        default:
          return "Thermometer" ;
      }
      break;
    }

    case OwwlDev_Cryo:
    {
      switch (devsubtype)
      {
        case OwwlCryo_Diode:
          return "Diode Thermometer" ;

        case OwwlCryo_Grt:
          return "GRT Thermometer" ;

        default:
          return "Cryogenic Thermometer" ;
      }
      break;
    }

    case OwwlDev_Barometer:
      return "Barometer" ;

    case OwwlDev_Humidity:
      return "Humidity" ;

    case OwwlDev_Wind:
      return "Wind Instrument" ;

    case OwwlDev_Rain:
      return "Rain Gauge" ;

    case OwwlDev_Gpc:
      return "General-Purpose Counter" ;

    case OwwlDev_Solar:
      return "Solar Sensor" ;

    case OwwlDev_Uvi:
      return "UV Index Sensor" ;

    case OwwlDev_Adc:
      return "ADC Channel";

    case OwwlDev_SoilMoisture:
      return "Soil Moisture";

    case OwwlDev_LeafWetness:
      return "Leaf Wetness";

    case OwwlDev_Psb:
      return "Polarization-Sensitive Bolometer" ;

    case OwwlDev_Diode:
      return "Diode Thermometer" ;

    case OwwlDev_Heater:
      return "Heater" ;

    case OwwlDev_Resistance:
      return "Resistance" ;

    case OwwlDev_Grt:
      return "Germanium Resistance Thermometer" ;
  }

  return "Unknown device type" ;
}

static char *
owwl_name_stem_short(int devtype, int devsubtype)
{
  switch (devtype)
  {
    case OwwlDev_Temperature:
    {
      switch (devsubtype)
      {
        case OwwlTemp_Humidity:
          return "Trh" ;

        case OwwlTemp_Barometer:
          return "Tb" ;

        case OwwlTemp_UVI:
          return "Tuv" ;

        case OwwlTemp_Solar:
          return "Tsol" ;

        case OwwlTemp_Adc:
          return "Tadc" ;

        case OwwlTemp_Thrmcpl:
          return "Ttc";

        case OwwlTemp_ThrmcplCJ:
          return "Tcj";

        case OwwlTemp_Soil:
          return "Tsoil";

        case OwwlTemp_Indoor:
          return "Tin";

        default:
          return "T" ;
      }
      break;
    }

    case OwwlDev_Cryo:
    {
      switch (devsubtype)
      {
        case OwwlCryo_Diode:
          return "Tdiode" ;

        case OwwlCryo_Grt:
          return "Tgrt" ;

        default:
          return "Tcryo" ;
      }
      break;
    }

    case OwwlDev_Barometer:
      return "BP" ;

    case OwwlDev_Humidity:
      return "RH" ;

    case OwwlDev_Wind:
      return "Wind" ;

    case OwwlDev_Rain:
      return "Rain" ;

    case OwwlDev_Gpc:
      return "GPC" ;

    case OwwlDev_Solar:
      return "Solar" ;

    case OwwlDev_Uvi:
      return "UVI" ;

    case OwwlDev_Adc:
      return "ADC" ;

    case OwwlDev_SoilMoisture:
      return "Moist";

    case OwwlDev_LeafWetness:
      return "Leaf";

    case OwwlDev_Psb:
      return "PSB" ;

    case OwwlDev_Diode:
      return "Diode" ;

    case OwwlDev_Heater:
      return "Heater" ;

    case OwwlDev_Resistance:
      return "R" ;

    case OwwlDev_Grt:
      return "Grt" ;
  }

  return "Unknown" ;
}

static int
inuse_off(owwl_conn *conn, owwl_data *data, void *user_data)
{
  data->in_use = 0 ;
  return 0 ; /* Ok */
}

unsigned int
owwl_data_arg_to_id(owwl_data *data, int arg)
{
  return (
    ((data->device_type) << 24) +
    ((data->device_subtype) << 16) +
    ((data->index) << 8) +
    arg
  ) ;
}

void
owwl_id_to_data_arg(owwl_conn *conn, unsigned int id, owwl_data **data, int *arg)
{
  unsigned int devtype, devsubtype, index ;

  /*g_assert((conn != NULL)) ;*/

  devtype    = (id & 0xff000000U) >> 24 ;
  devsubtype = (id & 0xff0000U)   >> 16 ;
  index      = (id & 0xff00U)     >>  8 ;
  *arg       = (id & 0xffU)             ;

  *data = owwl_find(conn, devtype, devsubtype, index) ;
}

/* Call a function for each data entry, in a set order */
/* The function should return non-zero to break out of the loop */

int
owwl_foreach_ordered(owwl_conn *conn, owwl_func func, void *user_data)
{
  int i ;
  unsigned int next_id = 0, potnext, id ;
  owwl_data *data = NULL ;

  do
  {
    data = NULL ;

    potnext = 0xffffffffU ;

    /* Loop over all data streams to find the next lowest id */
    for (i=0; i<conn->data_count; ++i)
    {
      id = owwl_data_arg_to_id(&(conn->data[i]), 0) ;
      if (id < next_id) continue ; /* Already covered this one */
      if (id < potnext) /* Could be the next one */
      {
        potnext = id ;
        data = &(conn->data[i]) ;
      }
    }

    if (!data) return 0 ; /* No more streams found */

    if (func(conn, data, user_data)) return -1 ;

    next_id = ++potnext ;
  } while (data) ;

  return 0 ; /* Ok */
}

/* Call a function for each data entry */
/* The function should return non-zero to break out of the loop */

int
owwl_foreach(owwl_conn *conn, owwl_func func, void *user_data)
{
  int i ;

  for (i=0; i<conn->data_count; ++i)
  {
    if (func(conn, &(conn->data[i]), user_data)) return -1 ;
  }

  return 0 ; /* Ok */
}

static void
set_none_in_use(owwl_conn *conn)
{
  owwl_foreach(conn, inuse_off, NULL) ;
}

/* Create a new owwl connexion context */

owwl_conn *
owwl_new_conn(int socket, void *user_data)
{
  owwl_conn *conn = NULL ;

  conn = (owwl_conn *) calloc(1, sizeof(owwl_conn)) ;

  if (conn)
  {
    conn->socket = socket ;
    conn->user_data = user_data ;
  }

  return conn ;
}

static int
owwl_decode(
  owwl_conn *conn,
  uint32_t head_word,
  uint32_t data_word[],
  uint32_t data_raw[],
  int *msg_type)
{
  long i ;
  int32_t *data_long ;
  float *data_float ;
  long length ;

  length = 0xffL & head_word ;

  i = (head_word & 0xff00) >> 8 ;

  data_long = (int32_t *) data_word ;
  data_float = (float *) data_word ;

#  ifdef OWWL_DEBUG
    printf("oww_trx_rx_decode(0x%x, [], *)\n", head_word) ;
#  endif
  switch (head_word & 0xff000000L)
  {
    case OWW_TRX_END:
#        ifdef OWWL_DEBUG
          printf("End of frame.\n") ;
#        endif
      /* We are done */
      return Owwl_Decode_Done ;

    case OWW_TRX_HEAD:
    {
      /* Check that the "Oww" identifier is present */
      if (strcmp((char *) data_raw, "Oww") != 0)
      {
#        ifdef OWWL_DEBUG
          printf("The remote data source is not using Oww protocol!\n") ;
#        endif
        return Owwl_Decode_Error ; /* Failed */
      }

      /* Get message type */
      *msg_type = (int) data_word[2] ;

      /* Get the protocol version, if given */
      if (length >= 4L)
      {
        owwl_remote_version = data_long[3] ;
      }
      else
      {
        owwl_remote_version = 0 ;
      }
#      ifdef OWWL_DEBUG
        printf("owwl_remote_version = %d\n",
          owwl_remote_version) ;
#      endif

      break ;
    }

    case OWW_TRX_LOC:
      /* Do we have latitude and longitude? */
#      ifdef OWWL_DEBUG
        printf("Received location record\n") ;
#      endif
      if (length >= 2L)
      {
        conn->latitude  = data_float[0] ;
        conn->longitude = data_float[1] ;
#       ifdef OWWL_DEBUG
         printf("Latitude = %f, Longitude = %f\n",
          conn->latitude, conn->longitude) ;
#       endif
      }
      break ;

    case OWW_TRX_UPDT:
      conn->interval = (int) data_word[2] ;
#      ifdef OWWL_DEBUG
        printf("Received time change (%d)\n", conn->interval) ;
#      endif
      break ;

    case OWW_TRX_TIME:
      conn->interval = (int) data_word[2] ;
      conn->data_time = (time_t) data_word[0] ;
      if (length >= 4L)
      {
        conn->data_time_usec = (long) data_word[3] ;
      }
      /* If we have times we have data */
#      ifdef OWWL_DEBUG
        printf("Received time data. interval = %d\n",
           conn->interval) ;
#      endif
      break ;

    case OWW_TRX_WIND:
    {
      owwl_data *wind ;

      wind = owwl_ensure_device(conn, OwwlDev_Wind, OwwlSubtype_Default, i) ;
      assert((wind != NULL)) ;

      wind->device_data.wind.bearing = 0.01F * (float) data_long[0] ;

      if (owwl_remote_version >= 100L)
      {
        wind->device_data.wind.speed = data_float[1] ;
        wind->device_data.wind.gust  = data_float[2] ;
      }
      else
      {
        wind->device_data.wind.speed = 0.001F * (float) data_long[1] ;
        wind->device_data.wind.gust  = 0.001F * (float) data_long[2] ;
      }
#      ifdef OWWL_DEBUG
        printf("Received wind data: %f, %f, %f\n",
          wind->device_data.wind.bearing,
          wind->device_data.wind.speed,
          wind->device_data.wind.gust) ;
#      endif
      break ;
    }

    case OWW_TRX_RAIN:
    {
      owwl_data *rain ;

      rain = owwl_ensure_device(conn, OwwlDev_Rain, OwwlSubtype_Default, i) ;
      assert((rain != NULL)) ;

      rain->device_data.rain.rain_count = data_word[0] ;
      rain->device_data.rain.rain_reset_time = (time_t) data_word[2] ;
      if (owwl_remote_version >= 100L)
      {
        rain->device_data.rain.rain_since_reset = 25.4 * data_float[1] ;
        rain->device_data.rain.rain_rate = 25.4 * data_float[3] ;
      }
      else
      {
        rain->device_data.rain.rain_since_reset = (float) data_long[1] * 0.01F ;
        rain->device_data.rain.rain_rate = (float) data_long[3] * 0.01F  ;
      }
#      ifdef OWWL_DEBUG
        printf("Received rain data: %f (%f) t=%d\n",
          rain->device_data.rain.rain_since_reset,
          rain->device_data.rain.rain_rate,
          rain->device_data.rain.rain_reset_time) ;
#      endif
      break ;
    }

    case OWW_TRX_T:
    {
      owwl_data *T ;
      int subtype ;

      switch (head_word & 0xff0000)
      {
        case OWW_TRX_SUB_TRH:
          subtype = OwwlTemp_Humidity ;
          break ;

        case OWW_TRX_SUB_TB:
          subtype = OwwlTemp_Barometer ;
          break ;

        case OWW_TRX_SUB_TUV:
          subtype = OwwlTemp_UVI ;
          break ;

        case OWW_TRX_SUB_TSOL:
          subtype = OwwlTemp_Solar ;
          break ;

        case OWW_TRX_SUB_TADC:
          subtype = OwwlTemp_Adc ;
          break ;

        case OWW_TRX_SUB_TTC:
          subtype = OwwlTemp_Thrmcpl ;
          break ;

        case OWW_TRX_SUB_TCJ:
          subtype = OwwlTemp_ThrmcplCJ ;
          break ;

        case OWW_TRX_SUB_TSOIL:
          subtype = OwwlTemp_Soil;
          break;

        case OWW_TRX_SUB_TINDOOR:
          subtype = OwwlTemp_Indoor;
          break;

        default:
          subtype = OwwlTemp_Thermometer ;
          break ;
      }

#      ifdef OWWL_DEBUG
        printf("Received T data: %f\n",
          data_float[0]) ;
#      endif

      T = owwl_ensure_device(conn, OwwlDev_Temperature, subtype, i) ;
      assert((T != NULL)) ;

      if (owwl_remote_version >= 100L)
        T->device_data.temperature.T = data_float[0] ;
      else
        T->device_data.temperature.T = (float) data_long[0] * 0.001F ;
      break ;
    }

    case OWW_TRX_RH:
    {
      owwl_data *rh ;
      rh = owwl_ensure_device(conn, OwwlDev_Humidity, OwwlSubtype_Default, i) ;
      assert((rh != NULL)) ;

      if (owwl_remote_version >= 100L)
      {
        rh->device_data.humidity.RH = data_float[0] ;
      }
      else
      {
        rh->device_data.humidity.RH = (float) data_long[0] * 0.001F ;
      }
#      ifdef OWWL_DEBUG
        printf("Received RH data: %f\n",
          rh->device_data.humidity.RH) ;
#      endif
      break ;
    }

    case OWW_TRX_BP:
    {
      owwl_data *bp ;
      bp = owwl_ensure_device(conn, OwwlDev_Barometer, OwwlSubtype_Default, i) ;
      assert((bp != NULL)) ;

     if (owwl_remote_version >= 100L)
      {
        bp->device_data.barom.bp = data_float[0] ;
      }
      else
      {
        bp->device_data.barom.bp = (float) data_long[0] * 0.001F ;
      }
#      ifdef OWWL_DEBUG
        printf("Received BP data: %f\n",
          bp->device_data.barom.bp) ;
#      endif
      break ;
    }

    case OWW_TRX_GPC:
    {
      owwl_data *gpc ;
      gpc = owwl_ensure_device(conn, OwwlDev_Gpc, OwwlSubtype_Default, i) ;
      assert((gpc != NULL)) ;

      gpc->device_data.gpc.gpc_calibrated = data_float[0] ;
      gpc->device_data.gpc.gpc_delta      = data_float[1] ;
      gpc->device_data.gpc.gpc_count      = data_word[2] ;
      gpc->device_data.gpc.gpc_rate       = data_float[3] ;
      gpc->device_data.gpc.gpc_reset_time = (time_t) data_word[4] ;
#      ifdef OWWL_DEBUG
        printf("Received GPC data: %f\n",
          gpc->device_data.gpc.gpc_calibrated) ;
#      endif
      break ;
    }

    case OWW_TRX_SOLAR:
    {
      owwl_data *sol ;
      sol = owwl_ensure_device(conn, OwwlDev_Solar, OwwlSubtype_Default, i) ;
      assert((sol != NULL)) ;

      sol->device_data.solar.solar = data_float[0] ;
#      ifdef OWWL_DEBUG
        printf("Received Solar data: %f\n",
          sol->device_data.solar.solar) ;
#      endif
      break ;
    }

    case OWW_TRX_UVI:
    {
      owwl_data *uvi ;
      uvi = owwl_ensure_device(conn, OwwlDev_Uvi, OwwlSubtype_Default, i) ;
      assert((uvi != NULL)) ;

      uvi->device_data.uvi.uvi = data_float[0] ;
#      ifdef OWWL_DEBUG
      printf("Received UVI data: %f\n",
    	  uvi->device_data.uvi.uvi) ;
#      endif
      break ;
    }

    case OWW_TRX_ADC:
    {
      owwl_data *adc ;
      adc = owwl_ensure_device(conn, OwwlDev_Adc, OwwlSubtype_Default, i) ;
      assert((adc != NULL)) ;

      adc->device_data.adc.V = 0.001 * data_float[0] ; /* Data broadcast as mV */
      adc->device_data.adc.I = data_float[1] ; /* Data broadcast as mV */
      adc->device_data.adc.Q = data_float[2] ;
#      ifdef OWWL_DEBUG
        printf("Received ADC data: %f, %f, %f\n",
        	adc->device_data.adc.V,
        	adc->device_data.adc.I,
        	adc->device_data.adc.Q
          ) ;
#      endif
      break ;
    }

    case OWW_TRX_SOIL:
    {
      owwl_data *soil ;
      soil = owwl_ensure_device(conn, OwwlDev_SoilMoisture, OwwlSubtype_Default, i) ;
      assert((soil != NULL)) ;

      soil->device_data.soilmoisture.soilmoisture = data_float[0] ;
#      ifdef OWWL_DEBUG
      printf("Received soil moisture data: %f\n",
    	  soil->device_data.soilmoisture.soilmoisture) ;
#      endif
      break ;
    }

    case OWW_TRX_LEAF:
    {
      owwl_data *leaf ;
      leaf = owwl_ensure_device(conn, OwwlDev_LeafWetness, OwwlSubtype_Default, i) ;
      assert((leaf != NULL)) ;

      leaf->device_data.leafwetness.leafwetness = data_float[0] ;
#      ifdef OWWL_DEBUG
      printf("Received leaf wetness data: %f\n",
    	  leaf->device_data.leafwetness.leafwetness) ;
#      endif
      break ;
    }

    case OWW_TRX_PSB:
    {
      owwl_data *psb ;
      psb = owwl_ensure_device(conn, OwwlDev_Psb, OwwlSubtype_Default, i) ;
      assert((psb != NULL)) ;

      psb->device_data.psb.psb_a = data_word[0] ;
      psb->device_data.psb.psb_b = data_word[1] ;
#      ifdef OWWL_DEBUG
        printf("Received PSB data: %d, %d\n",
          psb->device_data.psb.psb_a,
          psb->device_data.psb.psb_b) ;
#      endif
      break ;
    }

    case OWW_TRX_GRT:
    {
      owwl_data *grt ;
      grt = owwl_ensure_device(conn, OwwlDev_Grt, OwwlSubtype_Default, i) ;
      assert((grt != NULL)) ;

      grt->device_data.grt.V  = data_float[0] ;
      grt->device_data.grt.Vq = data_float[1] ;
      grt->device_data.grt.I  = data_float[2] ;
#      ifdef OWWL_DEBUG
        printf("Received GRT data: %g, %g, %g\n",
          grt->device_data.grt.V,
          grt->device_data.grt.Vq,
          grt->device_data.grt.I) ;
#      endif
      break ;
    }

    case OWW_TRX_DIODE:
    {
      owwl_data *diode ;
      diode = owwl_ensure_device(conn, OwwlDev_Diode, OwwlSubtype_Default, i) ;
      assert((diode != NULL)) ;

      diode->device_data.diode.V = data_float[0] ;
#      ifdef OWWL_DEBUG
        printf("Received Diode %d data: %f\n", i,
          diode->device_data.diode.V) ;
#      endif
      break ;
    }

    case OWW_TRX_HEATER:
    {
      owwl_data *heater ;
      heater = owwl_ensure_device(conn, OwwlDev_Heater, OwwlSubtype_Default, i) ;
      assert((heater != NULL)) ;

      heater->device_data.heater.V = data_float[0] ;
      
      if (length > 1)
      {
    	heater->device_data.heater.mA =  data_float[1] ;
    	heater->device_data.heater.haveI = 1;
      }
      else
      {
    	heater->device_data.heater.mA =  0.0F ;
    	heater->device_data.heater.haveI = 0;
      }

#      ifdef OWWL_DEBUG
        printf("Received Heater data: %f, %f\n",
          heater->device_data.heater.V, heater->device_data.heater.mA) ;
#      endif
      break ;
    }

    case OWW_TRX_RESISTANCE:
    {
      owwl_data *resistance ;
      resistance = owwl_ensure_device(conn, OwwlDev_Resistance, OwwlSubtype_Default, i) ;
      assert((resistance != NULL)) ;

      resistance->device_data.resistance.R = data_float[0] ;
#      ifdef OWWL_DEBUG
        printf("Received Resistance data: %f\n",
               resistance->device_data.resistance.R) ;
#      endif
      break ;
    }

    case OWW_TRX_CRYO:
    {
      owwl_data *T ;
      int subtype ;

      switch (head_word & 0xff0000)
      {
        case OWW_TRX_SUB_TCD:
          subtype = OwwlCryo_Diode ;
          break ;

        case OWW_TRX_SUB_TCG:
          subtype = OwwlCryo_Grt ;
          break ;

        default:
          subtype = OwwlCryo_Thermometer ;
          break ;
      }

#      ifdef OWWL_DEBUG
        printf("Received cryo T data: %f\n",
          data_float[0]) ;
#      endif

      T = owwl_ensure_device(conn, OwwlDev_Cryo, subtype, i) ;
      assert((T != NULL)) ;

      T->device_data.temperature.T = data_float[0] ;
      break ;
    }

    default:
    {
#      ifdef OWWL_DEBUG
        printf("Received unknown data type. Head word = &08x\n", head_word) ;
#      endif
        break;
    }


  }

  return Owwl_Decode_Not_Done ; /* Not done yet */
}

/* Try to read new data from server */

int
owwl_read(owwl_conn *conn)
{
  //uint32_t head_word = 0U ;
  uint32_t data_word[16] ;
  char *data_char;
  int done = Owwl_Decode_Not_Done, size, i ;
  int msg_type ;
  int words;

  assert(sizeof(float) == 4) ;

  data_char = (char *) &(conn->rd.data_word_n[0]);

  /* Try to read data from the server */

  if (conn->socket == INVALID_SOCKET)
  {
#ifdef OWWL_DEBUG
    printf("Owwl_Read_Error - Bad socket: %d", conn->socket) ;
#endif
	return Owwl_Read_Error ;
  }
  
  /* Ready to read remote data */
  /* There might be several chunks, so we loop */

  while (done == Owwl_Decode_Not_Done)
  {
    /* Using data_size to flag whether we are waiting for chunk header or data */
    if (conn->rd.data_size == 0)
    {
      data_char = (char *) conn->rd.data_word_n;
      while (conn->rd.head_bytes_read < 4) /* We need 4 bytes in total */
      {
        size = recv(conn->socket,
                    &data_char[conn->rd.head_bytes_read],
                    4 - conn->rd.head_bytes_read,
                    0) ;
        /*size = recv(conn->socket,
        (char *) &word_n,
        sizeof(uint32_t),
        0) ;*/

        /*werr(WERR_DEBUG0, "recv() [head] returned %d", size) ;*/

        if (size == SOCKET_ERROR)
        {
          /* Not ready or error */
          if (LASTSOCKERR == EWOULDBLOCK)
          {
            /* This is Ok - not ready yet - just return */
            return Owwl_Read_Again ;
          }

          /* Interrupted system call? */
          if (LASTSOCKERR == EINTR)
          {
            /* Slow call was interrupted - no problem - just return for another go */
            return Owwl_Read_Again ;
          }

#         ifdef OWWL_DEBUG
            printf("Owwl_Read_Error - recv() [head] failed: %d \"%s\"\n",
              LASTSOCKERR, strerror(LASTSOCKERR)) ;
#         endif
          return Owwl_Read_Error ; /* Connexion down */
        }

        else if (size == 0)
        {
#         ifdef OWWL_DEBUG
            printf("0 == recv() [head]\n") ;
#         endif /* ifdef OWWL_DEBUG */
          return Owwl_Read_Disconnect ;
        }
        
        conn->rd.head_bytes_read += size;
      }

      conn->rd.head_bytes_read = 0; /* Make ready for next header word */

      conn->rd.head_word = ntohl(conn->rd.data_word_n[0]) ;

#     ifdef OWWL_DEBUG
        printf("owwl_read - Received head word 0x%x\n", conn->rd.head_word);
#     endif /* ifdef OWWL_DEBUG */

      /* Now get the data */
      /* Size is encoded in the head word */

      conn->rd.data_to_get = conn->rd.data_size = sizeof(uint32_t) * (int) (conn->rd.head_word & 0xffL) ;
      data_char = (char *) conn->rd.data_word_n;
      if (conn->rd.data_size > sizeof(uint32_t) * 16)
      {
#     ifdef OWWL_DEBUG
        printf("Owwl_Read_Error - data_size = %d\n", conn->rd.data_size) ;
#     endif /* ifdef OWWL_DEBUG */
        return Owwl_Read_Error ;
      }
    }
    /* Will now have header, either for this or a previous call - proceed to read data */

    while(conn->rd.data_to_get > 0)
    {
      size = recv(conn->socket,
        /*(char *) &(conn->rd.data_word_n[conn->rd.data_size-conn->rd.data_to_get]),*/
        &data_char[conn->rd.data_size-conn->rd.data_to_get],
        conn->rd.data_to_get,
        0) ;
#     ifdef OWWL_DEBUG
        printf("Read %d bytes to data byte %d, size = %d\n", conn->rd.data_to_get, conn->rd.data_size-conn->rd.data_to_get, size) ;
#     endif /* ifdef OWWL_DEBUG */

      if (size == SOCKET_ERROR)
      {
#       ifdef OWWL_DEBUG
        printf("SOCKET_ERROR %d\n", LASTSOCKERR) ;
#       endif /* ifdef OWWL_DEBUG */
        
        /* Interrupted system call? */
        if (LASTSOCKERR == EINTR)
        {
#         ifdef OWWL_DEBUG
            printf("Owwl_Read_Again\n") ;
#         endif /* ifdef OWWL_DEBUG */
          /* Slow call was interrupted - no problem - just return for another go */
          return Owwl_Read_Again ;
        }

        if (LASTSOCKERR != EWOULDBLOCK)
        {
#         ifdef OWWL_DEBUG
            printf("Owwl_Read_Error - recv() failed: %d \"%s\"\n", LASTSOCKERR, strerror(LASTSOCKERR)) ;
#         endif /* ifdef OWWL_DEBUG */
          conn->rd.data_size = 0 ;
          return Owwl_Read_Error ; /* Connexion down */
        }
        
#       ifdef OWWL_DEBUG
        printf("Owwl_Read_Again\n") ;
#       endif /* ifdef OWWL_DEBUG */
        return Owwl_Read_Again ;
      }

      else if (size > 0)
      {
        conn->rd.data_to_get -= size /*/ sizeof(uint32_t)*/ ;
        /*break ; */
        //data_char += size;
        if (conn->rd.data_to_get > 0) continue ;
      }

      else if (size == 0)
      {
#        ifdef OWWL_DEBUG
          printf("0 == recv()\n") ;
#        endif /* ifdef OWWL_DEBUG */
        conn->rd.data_size = 0 ;
        return Owwl_Read_Disconnect ;
      }
    }

    if (conn->rd.data_size % 4 != 0)
    {
#     ifdef OWWL_DEBUG
      printf("Data size is not a whole number of words.\n") ;
#     endif /* ifdef OWWL_DEBUG */
      conn->rd.data_size = 0 ;
      return Owwl_Read_Error;
    }

    words = conn->rd.data_size / 4;
    
    for (i=0; i<words; ++i)
    {
      data_word[i] = ntohl(conn->rd.data_word_n[i]) ;
#     ifdef OWWL_DEBUG
        printf("DATA: %08x\n", data_word[i]) ;
#     endif /* ifdef OWWL_DEBUG */
    }

#   ifdef OWWL_DEBUG
      printf("Decode %08x\n", conn->rd.head_word) ;
#   endif /* ifdef OWWL_DEBUG */

    done = owwl_decode(
      conn, conn->rd.head_word, data_word, conn->rd.data_word_n, &msg_type) ;

    conn->rd.data_size = 0 ; /* Done with this chunk */
    conn->rd.head_word = 0U;

    if (done == Owwl_Decode_Error)
    {
#        ifdef OWWL_DEBUG
          printf("Owwl_Read_Error - Decode Error\n") ;
#        endif /* ifdef OWWL_DEBUG */
      conn->rd.data_size = 0 ;
      return Owwl_Read_Error ; /* Failure */
    }
  }

  /* Flag that now we have remote data */
  /*devices_remote_used = 1 ; */

  conn->rd.data_size = 0 ;
  return Owwl_Read_Read_And_Decoded ; /* Ok - done */
}

/* Free up finished-with connexion context */

int
owwl_free(owwl_conn *conn)
{
  int i ;

  if (!conn) return 0 ; /* Nothing to free */

  if ((conn->data_count > 0) && (conn->data)) free(conn->data) ;

  for (i=0; i<conn->nserver; ++i)
  {
    owwl_tx_server_kill_clients(&conn->server[i]) ;
    CLOSESOCK(conn->server[i].socket_listen) ;
    free(&conn->server[i]) ;
  }

  free(conn) ;

  return 0 ;
}

/* Get a name for a data entry */
/* Returns string - writes length */
/* Call with NULL buffer just to get length - no writing */

char *
owwl_name(owwl_data *data, char *buffer, int bufflen, int *length, int long_not_short)
{
  int stem_length, total_length ;
  char *name_stem ;
  char index_string[12] ;

  if (!data) return NULL ;

  name_stem = (long_not_short) ?
    owwl_name_stem_long(data->device_type, data->device_subtype) :
    owwl_name_stem_short(data->device_type, data->device_subtype) ;

  stem_length = strlen(name_stem) ;

  total_length = stem_length + sprintf(index_string, " %d", 1+data->index) ;

  /* If no buffer was provided, just return the length of the string */

  if (!buffer)
  {
    *length = total_length ;
    return NULL ;
  }

  /* We need space for a '\0' as well as total_length */

  if (total_length >= bufflen) return NULL ; /* Error */

  strcpy(buffer, name_stem) ;
  strcpy(&buffer[stem_length], index_string) ;
  *length = total_length ;

  return buffer ;
}

char *
owwl_arg_stem(enum owwl_device_type_enum device_type,
              int device_subtype,
              int arg)
{
  switch (device_type)
  {
    case OwwlDev_Temperature:
    case OwwlDev_Cryo:
    case OwwlDev_Barometer:
    case OwwlDev_Humidity:
    case OwwlDev_Solar:
    case OwwlDev_Uvi:
    case OwwlDev_Diode:
    case OwwlDev_Resistance:
    case OwwlDev_SoilMoisture:
    case OwwlDev_LeafWetness:
      return "" ;

    case OwwlDev_Heater:
    {
      switch (arg)
      {
      case 0:
    	return "V" ;
    	
      case 1:
    	return "I" ;
      }

      return "" ;
    }

    case OwwlDev_Wind:
    {
      switch (arg)
      {
        case 0:
          return "Speed" ;

        case 1:
          return "Gust" ;

        case 2:
          return "Bearing" ;
      }

      return "" ;
    }

    case OwwlDev_Rain:
    {
      switch (arg)
      {
        case 0:
          return "Rain" ;

        case 1:
          return "Since" ;

        case 2:
          return "Rate" ;

        case 3:
          return "Count" ;
      }

      return "" ;
    }

    case OwwlDev_Gpc:
    {
      switch (arg)
      {
        case 0:
          return "GPC" ;

        case 1:
          return "Since" ;

        case 2:
          return "Rate" ;

        case 3:
          return "Count" ;

        case 4:
          return "Delta" ;
      }

      return "" ;
    }

    case OwwlDev_Adc:
    {
      switch (arg)
      {
        case 0:
          return "Vdd" ;

        case 1:
          return "Vsns" ;

        case 2:
          return "Acc" ;
      }

      return "" ;
    }

    case OwwlDev_Psb:
    {
      switch (arg)
      {
        case 0:
          return "Ch. A" ;

        case 1:
          return "Ch. B" ;
      }

      return "" ;
    }

    case OwwlDev_Grt:
    {
      switch (arg)
      {
        case 0:
          return "V" ;

        case 1:
          return "Vq" ;

        case 2:
          return "I" ;
      }

      return "" ;
    }
    
    default:
      return "" ;

  }
  return "" ;
}

/* Get a name for a data entry argumet */
/* Returns string - writes length */
/* Call with NULL buffer just to get length - no writing */

char *
owwl_arg_name(owwl_data *data, char *buffer, int bufflen, int *length, int long_not_short, int arg)
{
  char *arg_stem ;
  int extra ;

  arg_stem = owwl_arg_stem(data->device_type, data->device_subtype, arg) ;
  extra = strlen(arg_stem) ;
  if (extra) ++ extra ;

  /* Just calculate length for buffer? */
  if (!buffer)
  {
    owwl_name(data, buffer, bufflen, length, long_not_short) ;

    if (extra) *length += extra ;
    return NULL ;
  }

  owwl_name(data, buffer, bufflen, length, long_not_short) ;

  if ((extra) && (bufflen >= *length + extra))
  {
    strcat(buffer, " ") ;
    strcat(buffer, arg_stem) ;
    *length += extra ;
  }

  return buffer ;
}

/* Find a device by type and index */

owwl_data *
owwl_find(owwl_conn *conn,
          enum owwl_device_type_enum device_type,
          int device_subtype,
          int index)
{
  int i ;

  assert((conn != NULL)) ;

  for (i=0; i< conn->data_count; ++i)
  {
    if
    (
      (conn->data[i].index == index)
      &&
      (
        (device_type == OwwlDev_Any) ||
        (conn->data[i].device_type == device_type)
      )
      &&
      (
        (device_subtype == OwwlSubtype_Any) ||
        (conn->data[i].device_subtype == device_subtype)
      )
    )
    {
      return &(conn->data[i]) ;
    }
  }

  return NULL ;
}

/* Find the first device of the given type */

owwl_data *
owwl_find_first(owwl_conn *conn,
                enum owwl_device_type_enum device_type,
                int device_subtype)
{
  int i ;

  assert((conn != NULL)) ;

  for (i=0; i< conn->data_count; ++i)
  {
    if
    (
      (
        (device_type == OwwlDev_Any) ||
        (conn->data[i].device_type == device_type)
      )
      &&
      (
        (device_subtype == OwwlSubtype_Any) ||
        (conn->data[i].device_subtype == device_subtype)
      )
    )
    {
      return &(conn->data[i]) ;
    }
  }

  return NULL ;
}

static int max_arg(owwl_data *data /*enum owwl_device_type_enum device_type*/)
{
  switch(data->device_type)
  {
    case OwwlDev_Temperature:
      return 0 ;

    case OwwlDev_Barometer:
      return 0 ;

    case OwwlDev_Humidity:
      return 0 ;

    case OwwlDev_Wind:
      return 2 ;

    case OwwlDev_Rain:
      return 3 ;

    case OwwlDev_Gpc:
      return 4 ;

    case OwwlDev_Solar:
      return 0 ;

    case OwwlDev_Uvi:
      return 0 ;

    case OwwlDev_Adc:
      return 2 ;

    case OwwlDev_SoilMoisture:
      return 0;

    case OwwlDev_LeafWetness:
      return 0;

    case OwwlDev_Psb:
      return 1 ;

    case OwwlDev_Diode:
      return 0 ;

    case OwwlDev_Grt:
      return 2 ;

    case OwwlDev_Cryo:
      return 0 ;

    case OwwlDev_Heater:
      return (data->device_data.heater.haveI==1) ? 1 : 0 ;

    case OwwlDev_Resistance:
      return 0 ;
      
    default:
      return 0 ;
  }

  return 0 ;
}

/* Get the next arg number */
int
owwl_next_arg(owwl_data *data, int arg)
{
  /*int max ; */

  assert((data->device_type >= 0)) ;
  assert((data->device_type < OwwlDev_Invalid)) ;

  if (++arg > max_arg(data))
    return -1 ;

  return arg ;
}

/* Add a word to the end of the buffer, expanding as necessary */
static int
owwl_buffadd(owwl_buffer *buff, uint32_t word)
{
  /* Do we need to expand memory allocation? */
  if (buff->next + 1 > buff->size)
  {
    /* Expand by BuffInc words */
    buff->buffer =
      (uint32_t *) realloc(buff->buffer, sizeof(uint32_t) * (OwwlBuffInc + buff->size)) ;
    if (buff->buffer == NULL)
    {
      /* Memory allocation failed */
      /*fprintf(stderr, "owwl_buffadd out of memory") ;*/
      return -1 ;
    }
    buff->size += OwwlBuffInc ;
  }

  buff->buffer[buff->next] = word ;
  ++buff->next ;
  return 0 ;
}

static int
owwl_buff_add_float(owwl_buffer *buff, float f)
{
  float lf ;
  lf = f ;

  if (0 != owwl_buffadd(buff, htonl(*((uint32_t *) &lf)))) return -1 ;
  return 0 ;
}

static int
owwl_buff_add_int(owwl_buffer *buff, int i)
{
  int li ;
  li = i ;

  if (0 != owwl_buffadd(buff, htonl(*((uint32_t *) &li)))) return -1 ;
  return 0 ;
}

/* Convert Unix time to local calendar time */
static time_t
owwl_convert_time(time_t t)
{
  struct tm *blt ;
  time_t ct ;

  /* Get broken-down UT */
  blt = gmtime(&t) ;

  /* Get Unix time from blt - but this presumes it is a local time
     so the correction is in the wrong direction */
  ct = mktime(blt) ;

  if (blt->tm_isdst)
    ct -= 3600 ;

  return (ct > t) ?
    t - (ct - t)  :
    t + (t - ct)  ;
}

static int
owwl_tx_add_data(owwl_conn *conn, owwl_data *data, void *user_data)
{
  owwl_buffer *buff ;

  buff = (owwl_buffer *) user_data ;

  assert(buff != NULL) ;


  /* How we add data to output buffer depends on data type */
  switch(data->device_type)
  {
  case OwwlDev_Any:
  case OwwlDev_Invalid:
	/* We should never see these - keep compiler happy */
	return -1;

    case OwwlDev_Temperature:
      switch(data->device_subtype)
      {
        case OwwlTemp_Humidity:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TRH +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Barometer:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TB +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_UVI:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TUV +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Solar:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TSOL +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Adc:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TADC +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Thrmcpl:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TTC +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_ThrmcplCJ:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TCJ +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Soil:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TSOIL +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlTemp_Indoor:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SUB_TINDOOR +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

       default:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_T + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      }
      if (0 != owwl_buff_add_float(buff, data->device_data.temperature.T)) return -1 ;
      break ;

    case OwwlDev_Barometer:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_BP + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.barom.bp)) return -1 ;
      break ;

    case OwwlDev_Humidity:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_RH + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.humidity.RH)) return -1 ;
      break ;

    case OwwlDev_Wind:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_WIND + OWW_TRX_SERIAL * data->index + 3L))) return -1 ;
      if (0 != owwl_buffadd(buff, htonl((long) (100.0F * data->device_data.wind.bearing)))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.wind.speed)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.wind.gust)) return -1 ;
      break ;

    case OwwlDev_Rain:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_RAIN + OWW_TRX_SERIAL * data->index + 4L))) return -1 ;
      if (0 != owwl_buffadd(buff, htonl(data->device_data.rain.rain_count))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.rain.rain_since_reset/25.4)) return -1 ;
      if (0 != owwl_buffadd(buff, htonl(data->device_data.rain.rain_reset_time))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.rain.rain_rate/25.4)) return -1 ;
      break ;

    case OwwlDev_Gpc:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_GPC + OWW_TRX_SERIAL * data->index + 5L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.gpc.gpc_calibrated)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.gpc.gpc_delta)) return -1 ;
      if (0 != owwl_buffadd(buff, htonl(data->device_data.gpc.gpc_count))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.gpc.gpc_rate)) return -1 ;
      if (0 != owwl_buffadd(buff, htonl(data->device_data.gpc.gpc_reset_time))) return -1 ;
      break ;

    case OwwlDev_Solar:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_SOLAR + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.solar.solar)) return -1 ;
      break ;

    case OwwlDev_Uvi:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_UVI + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.uvi.uvi)) return -1 ;
      break ;

    case OwwlDev_Adc:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_PSB + OWW_TRX_SERIAL * data->index + 3L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.adc.V * 1000.0F)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.adc.I)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.adc.Q)) return -1 ;
      break ;

    case OwwlDev_SoilMoisture:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_SOIL + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.soilmoisture.soilmoisture)) return -1 ;
      break ;

    case OwwlDev_LeafWetness:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_LEAF + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.leafwetness.leafwetness)) return -1 ;
      break ;

    case OwwlDev_Psb:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_PSB + OWW_TRX_SERIAL * data->index + 2L))) return -1 ;
      if (0 != owwl_buff_add_int(buff, data->device_data.psb.psb_a)) return -1 ;
      if (0 != owwl_buff_add_int(buff, data->device_data.psb.psb_b)) return -1 ;
      break ;

    case OwwlDev_Diode:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_DIODE + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.diode.V)) return -1 ;
      break ;

    case OwwlDev_Grt:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_GRT + OWW_TRX_SERIAL * data->index + 3L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.grt.V)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.grt.Vq)) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.grt.I)) return -1 ;
      break ;

    case OwwlDev_Cryo:
      switch(data->device_subtype)
      {
        case OwwlCryo_Diode:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_CRYO + OWW_TRX_SUB_TCD +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        case OwwlCryo_Grt:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_CRYO + OWW_TRX_SUB_TCG +
                  OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break ;

        default:
          if (0 != owwl_buffadd(buff,
            htonl(OWW_TRX_CRYO + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
          break;
      }
      if (0 != owwl_buff_add_float(buff, data->device_data.cryo.T)) return -1 ;
      break ;

    case OwwlDev_Heater:
      if (data->device_data.heater.haveI == 1)
      {
    	if (0 != owwl_buffadd(buff, htonl(OWW_TRX_HEATER + OWW_TRX_SERIAL * data->index + 2L))) return -1 ;
    	if (0 != owwl_buff_add_float(buff, data->device_data.heater.V)) return -1 ;
    	if (0 != owwl_buff_add_float(buff, data->device_data.heater.mA)) return -1 ;
      }
      else
      {
    	if (0 != owwl_buffadd(buff, htonl(OWW_TRX_HEATER + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
    	if (0 != owwl_buff_add_float(buff, data->device_data.heater.V)) return -1 ;
      }
      break ;

    case OwwlDev_Resistance:
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_RESISTANCE + OWW_TRX_SERIAL * data->index + 1L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, data->device_data.resistance.R)) return -1 ;
      break ;
  }

  return 0 ; /* Ok */
}

#ifdef _WIN32
void gettimeofday(struct timeval* pTv, void *pDummy)
{
    SYSTEMTIME sysTime;
    FILETIME fileTime;	/* 100ns == 1 */
    LARGE_INTEGER i;

    GetSystemTime(&sysTime);
    SystemTimeToFileTime(&sysTime, &fileTime);
    /* Documented as the way to get a 64 bit from a FILETIME. */
    memcpy(&i, &fileTime, sizeof(LARGE_INTEGER));

    pTv->tv_sec = i.QuadPart / 10000000; /*10e7*/
    pTv->tv_usec = (i.QuadPart / 10) % 1000000; /*10e6*/

}
#else

#ifndef HAVE_GETTIMEOFDAY
void gettimeofday(struct timeval* pTv, void *pDummy)
{
  struct timespec ts;

  clock_gettime(CLOCK_REALTIME, &ts);

  pTv->tv_sec  = ts.tv_sec;
  pTv->tv_usec = ts.tv_nsec / 1000;
}
#endif /* HAVE_GETTIMEOFDAY */
#endif /* _WIN32 */

/* Add a timestamp to the conn */

void
owwl_tx_add_timestamp(owwl_conn *conn)
{
  struct timeval tv ;

  gettimeofday(&tv, NULL) ;

  conn->data_time = tv.tv_sec ;
  conn->data_time_usec = tv.tv_usec ;
}

static int
owwl_tx_build_data(owwl_conn *conn, owwl_buffer *buff)
{
  /* Add common stuff */
  /* Times */

  if (0 != owwl_buffadd(buff, htonl(OWW_TRX_TIME + 4L))) return -1 ;
  if (0 != owwl_buffadd(buff, htonl(conn->data_time))) return -1 ;
  if (0 != owwl_buffadd(buff, htonl(owwl_convert_time(conn->data_time)))) return -1 ;
  if (0 != owwl_buffadd(buff, htonl(conn->interval))) return -1 ;
  if (0 != owwl_buffadd(buff, 0L)) return -1 ; /* Not using microsecond slot */

  /* Add data from each data channel */

  owwl_foreach_ordered(conn, owwl_tx_add_data, (void *) buff) ;

  return 0 ;
}

static int
owwl_tx_build_head(owwl_conn *conn, int msg_type, owwl_buffer *buff)
{
  char ident[] = "Oww" ;
  int *identi ;

  identi = (int *) ident ;

  /* Construct header */

  /*oww_trx_bufexp(5) ; */

  if (0 != owwl_buffadd(buff, htonl(OWW_TRX_HEAD + 4L))) return -1 ;
  if (0 != owwl_buffadd(buff, *identi)) return -1 ;
  if (0 != owwl_buffadd(buff, 0)) return -1 ; /* This will hold the size */
  if (0 != owwl_buffadd(buff, htonl(msg_type & 0xffL))) return -1 ;
  if (0 != owwl_buffadd(buff, htonl(OWW_PROTO_VERSION))) return -1 ;

  return 0 ; /* Ok */
}

/* Fill a buffer with data for transmission */
int
owwl_tx_build(owwl_conn *conn, int msg_type, owwl_buffer *buff)
{
  buff->next = 0 ;

  assert(sizeof(float) == 4) ;

  /* Construct header */

  if (0 != owwl_tx_build_head(conn, msg_type, buff)) return -1 ;

  switch (msg_type)
  {
    case OWW_TRX_MSG_ANNOUN:
      /* Initial announcement - send location */
      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_LOC + 2L))) return -1 ;
      if (0 != owwl_buff_add_float(buff, conn->latitude)) return -1 ;
      if (0 != owwl_buff_add_float(buff, conn->longitude)) return -1 ;

    /* Fall through for normal WS data */

    case OWW_TRX_MSG_WSDATA:
    {
      /* Normal data update */
      if (0 != owwl_tx_build_data(conn, buff)) return -1 ;
      break ;
    }

    case OWW_TRX_MSG_UPDT:
      /* Change in update interval */
      /* Times */

      if (0 != owwl_buffadd(buff, htonl(OWW_TRX_UPDT + 1L))) return -1 ;
      if (0 != owwl_buffadd(buff, htonl(conn->interval))) return -1 ;
      break ;

    case OWW_TRX_MSG_MORT:
      /* Server is dying - no data */
      break ;

  }

  /* End */

  if (0 != owwl_buffadd(buff, htonl(OWW_TRX_END))) return -1 ;

  /* Set size in oww_trx_buffer[1] */
  assert(buff->size >= 3) ;
  buff->buffer[2] = htonl(buff->next * (long) sizeof(uint32_t)) ;

  return 0 ; /* Ok */
}

/* Update or add data values in a data stream */
int
owwl_data_in(
  owwl_conn *conn,
  enum owwl_device_type_enum device_type,
  int device_subtype,
  int index,
  ...)
{
  va_list ap ;
  owwl_data *data ;

  va_start(ap, index) ;

  data = owwl_ensure_device(conn, device_type, device_subtype, index) ;
  if (data == NULL) return -1 ; /* Error - could get data pointer */

  switch (device_type)
  {
    case OwwlDev_Temperature: /* Temperature takes one float arg */
      data->device_data.temperature.T = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Barometer: /* Barometer takes one float arg */
      data->device_data.barom.bp = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Humidity: /* Humidity takes one float arg */
      data->device_data.humidity.RH = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Wind: /* Wind takes three floats: speed, gust, bearing */
      data->device_data.wind.speed   = (float) va_arg(ap, double) ;
      data->device_data.wind.gust    = (float) va_arg(ap, double) ;
      data->device_data.wind.bearing = (float) va_arg(ap, double) ;
     break ;

    case OwwlDev_Rain: /* Rain takes: count, reset time, rain, rain rate  */
      data->device_data.rain.rain_count       = va_arg(ap, int) ;
      data->device_data.rain.rain_reset_time  = va_arg(ap, time_t) ;
      data->device_data.rain.rain_since_reset = (float) va_arg(ap, double) ;
      data->device_data.rain.rain_rate        = (float) va_arg(ap, double) ;
     break ;

    case OwwlDev_Gpc: /* GPC takes: count, reset time, calibrated, rate, delta  */
      data->device_data.gpc.gpc_count      = va_arg(ap, int) ;
      data->device_data.gpc.gpc_reset_time = va_arg(ap, time_t) ;
      data->device_data.gpc.gpc_calibrated = (float) va_arg(ap, double) ;
      data->device_data.gpc.gpc_rate       = (float) va_arg(ap, double) ;
      data->device_data.gpc.gpc_delta      = (float) va_arg(ap, double) ;
     break ;

    case OwwlDev_Solar: /* Solar takes one float arg */
      data->device_data.solar.solar = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Uvi: /* Uvi takes one float arg */
      data->device_data.uvi.uvi = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Adc: /* ADC takes three float args */
      data->device_data.adc.V = (float) va_arg(ap, double) ;
      data->device_data.adc.I = (float) va_arg(ap, double) ;
      data->device_data.adc.Q = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_SoilMoisture: /* Soil moisture takes one float arg */
      data->device_data.soilmoisture.soilmoisture = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_LeafWetness: /* Leaf wetness takes one float arg */
      data->device_data.leafwetness.leafwetness = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Psb: /* PSB takes two int args */
      data->device_data.psb.psb_a = va_arg(ap, int) ;
      data->device_data.psb.psb_b = va_arg(ap, int) ;
      break ;

    case OwwlDev_Diode: /* Diode takes one float arg */
      data->device_data.diode.V = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Grt: /* Grt takes three float args */
      data->device_data.grt.V =  (float) va_arg(ap, double) ;
      data->device_data.grt.Vq = (float) va_arg(ap, double) ;
      data->device_data.grt.I =  (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Cryo: /* Calibrated cryogenic temperature takes one float arg */
      data->device_data.temperature.T = (float) va_arg(ap, double) ;
      break ;

    case OwwlDev_Heater: /* Heater takes two float args */
      data->device_data.heater.V = (float) va_arg(ap, double) ;
      data->device_data.heater.mA = (float) va_arg(ap, double) ;
      data->device_data.heater.haveI = 1;
      break ;

    case OwwlDev_Resistance: /* Resistance takes one float arg */
      data->device_data.resistance.R = (float) va_arg(ap, double) ;
      break ;

    default:
      va_end(ap) ;
      return -2 ; /* Bad device_type */
  }

  va_end(ap) ;

  return 0 ; /* Ok */
}

/* Close a client socket and unlink struct from list */
static void
owwl_tx_server_kill_client(owwl_server *server, owwl_client *client)
{
  CLOSESOCK(client->socket) ;

  if (client->prev != NULL)
    (client->prev)->next = client->next ;
  else
    server->first_client = client->next ;

  if (client->next != NULL)
    (client->next)->prev = client->prev ;

  free((void *) client) ;
}

static void
owwl_tx_server_kill_clients(owwl_server *server)
{
  owwl_client *client ;

  client = server->first_client ;

  while (client != NULL)
    owwl_tx_server_kill_client(server, client) ;
}

/* Create a new client */
static owwl_client *
owwl_tx_new_client(owwl_server *server)
{
  owwl_client *new_client ;

  new_client = (owwl_client *) malloc(sizeof(owwl_client)) ;

  if (new_client != NULL)
  {
    if (server->first_client != NULL)
      (server->first_client)->prev = new_client ;

    new_client->next = server->first_client ;
    new_client->prev = NULL ;
    server->first_client = new_client ;
  }

  return new_client ;
}

/* Poll clients of server to flush rx data and collect pending close */
static int
owwl_tx_server_poll_clients(owwl_server *server)
{
  int received ;
  char buffer[64];
  owwl_client *client, *next_client ;

  client = server->first_client ;

  /* Loop over all clients */
  while (client != NULL)
  {
    next_client = client->next ;

    /* Read in incoming data and dump */
    do
      received = recv (client->socket, buffer, 64, 0) ;
    while (received > 0) ;

    if (received == 0) /* Client is closing connexion */
      owwl_tx_server_kill_client(server, client) ;

    /* On to next client or end */
    client = next_client ;
  }

  return 0 ;
}

/* Send contents of buffer to a client */
static int
owwl_tx_to_client(owwl_server *server, owwl_client *client, owwl_buffer *buff)
{
  int sent, len ;

  if (client->socket > 0)
  {
    len = sizeof(uint32_t) * buff->next ;
    /* This should be a valid socket */

    sent = send(client->socket, (void *) buff->buffer, len, MSG_DONTWAIT);
    if (len != sent)
    {
      /* Just ignore calls that would block */
      if ((sent < 0) && (LASTSOCKERR == EWOULDBLOCK))
      {
        /* perror("Error sending to client") ;*/
        return -2 ;
      }

      /* Some kind of error - client disconnected probably */
      /*perror("Send to client failed. Closing connexion.") ;*/
      owwl_tx_server_kill_client(server, client) ;
      return -1 ;
    }
  }
  return 0 ; /* Ok */
}

/* Send contents of buffer to all clients */
int
owwl_tx(owwl_conn *conn, owwl_buffer *buff)
{
  owwl_client *client ;
  int i, rval = 0 ;

  for (i=0; i<conn->nserver; ++i)
  {
    client = conn->server[i].first_client ;

    while (client != NULL)
    {
      if (0 != owwl_tx_to_client(&(conn->server[i]), client, buff)) rval = -1 ;
      client = client->next ;
    }
  }
  return rval ;
}

/* A listing socket is ready with a connexion request - accept new client */
static int
owwl_tx_accept(owwl_conn *conn, owwl_server *server)
{
  owwl_client *new_client ;
  int s ;
  uint32_t a ;

  s = accept(server->socket_listen, NULL, NULL);

	a = 1L;
	IOCTL (s, FIONBIO, &a);

  if (s == -1)
  {
    /* Error */
    /*perror("Error accepting connexion") ;*/
    /* Interrupted system call */
    if (LASTSOCKERR == EINTR)
    {
      /* Slow call was interrupted - no problem - just return for another go */
      return Owwl_Read_Again ;
    }

    return -1 ;
  }

  /*printf("New client connexion accepted\n") ;*/

  /* Add this socket to the list of clients */
  new_client = owwl_tx_new_client(server) ;

  if (new_client != NULL)
  {
    owwl_buffer buff = Owwl_Buffer_Init ;

    new_client->socket = s ;

    /* Build announcement message */
    if (0 != owwl_tx_build(conn, OWW_TRX_MSG_ANNOUN, &buff)) return -1 ;

    /* Send announcement message to this client */
    if (0 != owwl_tx_to_client(server, new_client, &buff)) return -1 ;

    /* Free memory allocated by owwl_tx_build() */
    free((void *) buff.buffer) ;
  }
  return 0 ;
}

/* Poll for incoming data and connexion requests */
int
owwl_tx_poll_servers(owwl_conn *conn)
{
#if defined(_WIN32) || defined(VXW)
  /* Use select() polling for Windows and VXWorks */

  fd_set fds ;
  int i, poll_result ;
  int n = 0 ; /* Highest listening socket */
  struct timeval tv = {0, 100} ;

  FD_ZERO(&fds) ;

  for (i=0; i<conn->nserver; ++i)
  {
    /* Poll listening sockets for incoming data */
    owwl_tx_server_poll_clients(&conn->server[i]) ;
    FD_SET(conn->server[i].socket_listen, &fds) ;
    if (conn->server[i].socket_listen > n)
      n = conn->server[i].socket_listen ;
  }

  poll_result = select(n+1, &fds, NULL, NULL, &tv) ;

  if (poll_result == SOCKET_ERROR)
  {
    /* Error */
    /*perror("Error polling server listing sockets") ;*/
    return -1 ;
  }

  if (poll_result == 0) return 0 ; /* No connexion requests */

  /* poll_result > 0 => new client */

  for (i=0; i<conn->nserver; ++i)
  {
    /* Accept each incoming request */
    if (FD_ISSET(conn->server[i].socket_listen, &fds))
    {
      if (0 != owwl_tx_accept(conn, &conn->server[i])) return -1 ;
    }
  }

#else

  int i, poll_result ;
  struct pollfd *fds ;

  fds = (struct pollfd *) calloc(conn->nserver, sizeof(struct pollfd)) ;
  if (fds == NULL)
  {
    /*fprintf(stderr, "Not enough memory for server poll\n") ;*/
    return -1 ;
  }

  for (i=0; i<conn->nserver; ++i)
  {
    /* Poll listening sockets for incoming data */
    owwl_tx_server_poll_clients(&conn->server[i]) ;
    fds[i].fd = conn->server[i].socket_listen ;
    fds[i].events = POLLIN ; /* Data to read => accept ready */
  }

  poll_result = poll(fds, conn->nserver, 50) ;

  if (poll_result == -1)
  {
    /* Error */
    /*perror("Error polling server listing sockets") ;*/
    free(fds) ;
    return -1 ;
  }

  if (poll_result == 0) 
  {
    free(fds) ;
    return 0 ; /* No connexion requests */
  }

  /* poll_result > 0 => new client */

  for (i=0; i<conn->nserver; ++i)
  {
    /* Accept each incoming request */
    if ((fds[i].revents | POLLIN) != 0)
    {
      if (0 != owwl_tx_accept(conn, &conn->server[i]))
      {
        free(fds) ;
        return -1 ;
      }
    }
  }
  free(fds) ;
#endif
  return 0 ;
}

/* Create a new server from a listening socket */
int
owwl_tx_new_server(owwl_conn *conn, int socket)
{
  conn->server = (owwl_server *) realloc((void *) conn->server,
    (1 + conn->nserver) * sizeof(owwl_server)) ;
  if (conn->server == NULL)
  {
    /*fprintf(stderr, "No memory for server\n") ;*/
    return -1 ;
  }

  conn->server[conn->nserver].socket_listen = socket ;
  conn->server[conn->nserver].first_client = NULL ;
  ++conn->nserver ;
  return 0 ; /* Ok */
}
